
"""
py-file-conf is a Python framework for flow-based programming and managing configuration. To use it, you register
your main functions and classes, and config files are created automatically for them. It provides a way to run these functions
individually, in a list, in sections, or a combination thereof. Configuration can be 
dynamically updated, enabling powerful scripting.
"""
from pyfileconf.plugin.impl import hookimpl
from pyfileconf.main import PipelineManager
from pyfileconf.selector.models.selector import Selector
from pyfileconf.iterate import IterativeRunner
from pyfileconf.batch import BatchUpdater
from pyfileconf.plugin.manager_utils import reset_plugins, remove_default_plugins
from pyfileconf.context import context
from pyfileconf.selector.models.itemview import is_item_view
